* SuperBASIC FPU enhancement routines, v 1.7
* Simon N Goodwin Amiga Qdos 18th March 1997
* Assemble with GWASS for FPU opcode support
*
* Replacement for SIN, COS, ASIN, ACOS, LN, LOG10, EXP, SQRT,
* TAN, COT, ATAN, ACOT. Requires >=68020 with a Motorola FPU.
*
* History:
*
* Version 1.7 - April 1997
*
*           CALL version is now compatible with 64 byte
*            FPSP headers (for FPSAVE 1.17 and later)
*
* Version 1.6 - March 1997
*
*           COT(0) returns overflow (like SMS/QE) rather
*             than 1 (like Qdos, which is surely wrong!)
*
* Version 1.5 - March 1997
*
*           68060 'fix' no longer confuses all other CPUs!
*           Relaxed check as only vectors <64 are required.
*           Source code tidied and further comments added.
*
* Version 1.4 - March 1997
*
*           68060 initialisation corrected
*           IEEE mantissa $FFFFFFFFxxxxxxxx is
*             rounded correctly (thanks George)
*
* Version 1.3 - January 1997
*
*           ACOT added and tested
*           Link to resident FPSP altered
*           COT bug fixed - thanks George
*           EXP range increased to +/-1024
*           FNULL added - thanks Thierry
*
* Version 1.2i - January 1997
*
*           Branch table checking now works as planned
*           COT still seems to have an overflow bug
*
* Version 1.2h - January 1997
*
*           Unreported Overflow bugs fixed (missing RTS)
*           FPSAVE link option implemented (set DEBUG=0)
*           EXP supports same input range as SuperBASIC
*           ASIN and ACOS check and optimise parameters
*
* Version 1.2e - December 1996
*
*           First test release for George Gwilt and Thierry Godefroy
*
* Configuration constants - tweak these for test versions of the code

DEBUG       equ         0                       Set for CALL linkage
FNAMES      equ         1                       Use F extension prefix?

bias        equ         14334                   Exponent difference
** 1.7 ** Head equ      40                      Offset to branch table

* Vector address offsets in George Gwilt's remix of Motorola's FPSP

fSinX       equ         $1B0
fASinX      equ         $30
fCosX       equ         $78
fACosX      equ         $18
fExpX       equ         $A8                     E ^ X = EXP(X)
fLnX        equ         $138
fLog10X     equ         $108
fTanX       equ         $1F8
fATanX      equ         $48

* Short branch optimisation macros - MUST not be used to skip branches
* on 68060. These use TRAPF, taking the subsequent word as its (dummy)
* parameter, for speed. They only work on 68020 or later processors.

SkipWord    macro
            dc.w        $51FA                   ;TRAPF.W opcode
            endm

SkipLong    macro
            dc.w        $51FB                   ;TRAPF.L opcode
            endm

* These are Motorola-recommended quick ways to skip a word without
* upsetting the processor pipelines. Poor 680XX emulators might be
* more likely to work with simple BRA.S instructions, but should I
* penalise users of real 680XXs for them? Change it, if you like.

            if          DEBUG=0

* Link to George Gwilt's FPSAVE system unless debugging FPU.
* Reports 'not implemented' if FPSAVE is not already loaded,
* or its memory area does not start with 100 branch opcodes.

start       moveq       #0,d0                   MT.INF
            trap        #1                      A0 -> system variables
            move.l      $D8(a0),d1              FPSP pointer
            beq.s       give_up                 FPSP is not loaded!

            movea.l     d1,a1                   A1 -> Access routine
            suba.l      a0,a0                   No user exceptions
            jsr         (a1)                    Find FPSP start

            else

* CALL this routine with address of FPSP library in second parameter
*  if you want to debug without linking the FPSP into the scheduler

start       
** 1.7 **   add.l       #Head,d1                Skip to branches
            move.l      d1,a1
            adda.l      (a1),a1                 **1.7** Use offset

            end_if                              DEBUG

**1.5** Check that the FPSP starts with a plausible jump table

            move.l      a1,d1                   Keep two copies
            moveq       #96,d0                  BRA opcode byte
            moveq       #63,d2                  Test 64 entries
check       cmp.b       (a1),d0                 Is this a branch?
            addq.l      #8,a1                   Advance to next
            dbne        d2,check
            beq.s       set_vectors             Library not found?

* There should be lots of branch op-bytes at eight byte intervals
* at the address the FPSP returns. Otherwise we can't rely on the
* availability of vital FPSP vectors, so FPUfns must not link in.

give_up     moveq       #-19,d0                 Not implemented
            rts

* Point our vectors into the FPSP (extra branches seem superfluous)

set_vectors lea.l       Vectors,a1
            moveq       #8,d2                   (Vectors-Vector_End)/4-1
            tst.w       (a1)                    Has it moved?
            beq.s       set_one

            moveq       #-9,d0                  In use - re-CALLed!
            rts

set_one     add.l       d1,(a1)+                Add base to offset
            dbra        d2,set_one

* Link the functions to SuperBASIC

            lea.l       define,a1
            movea.w     ($110).w,a2
            jsr         (a2)

            suba.l      a0,a0                   Channel #0 or #1
            lea.l       message,a1
            movea.w     ($D0).w,a2              UT.MTEXT vector
            jmp         (a2)

* Start-up message - helps to keep the many versions distinct!

message     dc.w        36                      MODE 8 TV limit = 36

            dc.b        "Simon Goodwin's FPU FUNCTIONS v 1.7",10

* Non-re-entrant data area - no problem if repeat calls are trapped

            cnop        0,4
Vectors
Sine        dc.l        fSinX
ArcSine     dc.l        fASinX
CoSine      dc.l        fCosX
ArcCoSine   dc.l        fACosX
Exp         dc.l        fExpX
Ln          dc.l        fLnX
Log10       dc.l        fLog10X                 SQRT uses no vector
Tangent     dc.l        fTanX                   COT shares this one
ArcTangent  dc.l        fATanX                  ACOT re-uses this
Vector_End

* There is some duplication in the code that follows, as the aim is
* to produce fast code and the overhead of the FPSP dwarfs the size
* of these routines. For speed, most 0 parameters avoid FPSP calls,
* and manipulation of the IEEE sign is avoided if arguments permit.
*
* To do:
*
* 1) Trap overflow in any remaining case and report to SuperBASIC
* 2) Test accuracy and speed; compare with original Sinclair code
*
* The first half dozen testers have only reported one problem in
* the second category (fixed in 1.4) and none in the first, as of
* version 1.3; there are small discrepancies between Qdos and FPU
* results caused by discontinuities and argument reduction. These
* functions are sure to have other subtleties yet unfathomed; your
* comments are very welcome. Please EMAIL: simon@studio.woden.com
*
*******************************************************************
*
* Replacement SuperBASIC ARCCOTANGENT function using FPU IEEE maths

facot       move.l      ArcTangent,a4           FPSP vector
            movea.w     ($114).w,a2             Get FP parameter
            jsr         (a2)
            bne         faulty

            subq.w      #1,d3                   One at a time
            bne         bad_param

            move.w      0(a1,a6.l),d0
            beq.s       piby2                   ACOT(0)=PI/2

* Convert mantissa at 2(a1,a6.l) and exponent in D0 to IEEE format

            addi.w      #bias,d0                Bias exponent
            clr.l       -(a7)
            subq.l      #4,a7                   No need to clear
            clr.l       -(a7)
            move.w      d0,(a7)                 Plug in exponent
            move.l      2(a1,a6.l),d0
            bpl.s       move_left

            bset        #7,(a7)                 Adjust sign
            neg.l       d0
            bpl.s       move_left               Already normalised?

            addq.w      #1,(a7)                 Scale exponent up
            SkipWord

move_left   add.l       d0,d0                   Normalise mantissa
            move.l      d0,4(a7)

* Since we reckon ACOT(X) = ATAN(1/X), work out the reciprocal of X

            fmove.x     (a7),fp1
            fmove.s     #!1,fp0
            fdiv.x      fp1,fp0                 Reciprocate!
            move.l      ArcTangent,a4           Method vector
            fmove.x     fp0,(a7)
            bra         call_fpsp

*******************************************************************
*
* Replacement SuperBASIC ARCCOSINE function using FPU IEEE maths

farccosine  move.l      ArcCosine,a4            FPSP vector
            movea.w     ($114).w,a2             Get FP parameter
            jsr         (a2)
            bne         faulty

            subq.w      #1,d3                   One at a time
            bne         bad_param

            move.w      0(a1,a6.l),d0
            beq.s       piby2                   ACOS(0)=PI/2

            cmpi.w      #$0801,d0
            bcs         convert                 In range -1..+1
            bne.s       undefined               Beyond that range

* Exponent =$0801; check for the special case of exactly +1

            move.l      2(a1,a6.l),d1           Get mantissa
            cmpi.l      #$40000000,d1
            bne.s       undefined                ACOS(1)=0

zerolong    clr.l       0(a1,a6.l)              Last word already=0
            bra         fine

* Return PI/2 (exponent may already be $0801) in Qdos form

piby2       move.w      #$0801,0(a1,a6.l)
pi_mantissa move.l      #$6487ED51,2(a1,a6.l)
            bra         fine

*******************************************************************
*
* Replacement SuperBASIC ARCSINE function uses IEEE maths in an FPU

farcsine    move.l      ArcSine,a4
            movea.w     ($114).w,a2             Get FP parameter
            jsr         (a2)
            bne.s       faulty

            subq.w      #1,d3                   One at a time
            bne         bad_param

            move.w      0(a1,a6.l),d0
            beq.s       fine                    ASIN(0)=0

            cmpi.w      #$0801,d0
            bcs         convert                 In range -1..+1
            bne.s       undefined               Beyond that range

* Exponent =$0801; check for the special case of exactly +1

            move.l      2(a1,a6.l),d1           Get mantissa
            cmpi.l      #$40000000,d1
            beq.s       pi_mantissa             ASIN(1)=PI/2

undefined   moveq       #-18,d0
faulty      rts

*******************************************************************
*
* Replacement SuperBASIC COTANGENT function for maths co-processor
*
* This one is non-trivial because it calls the FPSP Tangent routine
* then fiddles with the result on the stack to make it a CoTangent.
*
* The code that does the work, with an indirect call to the FPSP

CoTangent   move.l      Tangent,a2
            move.l      (a7)+,a4                Return for later
            jsr         (a2)                    Hope A4 survives

* Error trapping may be needed here in case of overflow
*
*           tst.l       d0
*           bmi.s       snag

            fmove.x     fp0,fp1
            fmove.s     #!1,fp0
            fdiv.x      fp1,fp0                 Reciprocal

* More error trapping needed here (or before) in case of overflow?
*
*           tst.l       d0
*           bmi.s       snag                    One at a time
*           bne.s       bad_param               FIX! for v1.3

            jmp         (a4)                    Tally ho

fcotangent  lea.l       CoTangent,a4
            move.w      ($114).w,a2             Qdos get FP vector
            jsr         (a2)
            bne.s       snag                    Error code in D0
            
            subq.w      #1,d3                   One parameter only?
            bne.s       bad_param

            move.w      0(a1,a6.l),d0           Get exponent
            beq.s       undefined               Out of range?
            bra.s       convert                 Not 0 special case

*******************************************************************
*
* Replacement SuperBASIC COSINE function using IEEE format maths
            
fcosine     move.l      Cosine,a4
            movea.w     ($114).w,a2             Get FP parameter
            jsr         (a2)
            bne.s       snag

            subq.w      #1,d3                   One at a time
            bne.s       bad_param

            move.w      0(a1,a6.l),d0
            bne.s       convert                 COS(0)=1

one         move.w      #2049,0(a1,a6.l)        Qdos exponent of 1
            move.l      #$40000000,2(a1,a6.l)   A concession to Hex

fine        moveq       #2,d4                   Qdos FP result
            moveq       #0,d0
            rts

bad_param   moveq       #-15,d0
snag        rts

*******************************************************************
*
* Replacement SuperBASIC TAN function using the maths co-processor

ftangent    move.l      Tangent,a4
            bra.s       fskip0                  Note TAN(0)=0

* In theory this needs error trapping or some way to propagate
* overflows out, but in practice they do not occur because the
* discontinuity at TAN(PI/2) and multiples thereof is BETWEEN
* two Qdos floating-point values, and the correct result for
* both those variables is large but not infinite. This is the
* same behaviour as for "JS" Sinclair SuperBASIC which returns
* 1.6455713E10; FPUFNs with FPSP_040 returns 1.6455216E10 which
* is not identical (there are other small discrepancies) but
* very close - I'm not sure which is closer to the 'correct'
* answer in this case, but infinity is definitely way off, so
* the SMS/QE result of an overflow error appears incorrect.
*
*******************************************************************
*
* Replacement SuperBASIC ATAN function using IEEE mathematics

farctangent move.l      ArcTangent,a4           Note ATAN(0)=0

*           bra.s       fskip0
            SkipLong

*******************************************************************
*
* Replacement SuperBASIC SINE function for maths co-processor

fsine       move.l      Sine,a4                 PC relative only
fskip0      move.w      ($114).w,a2             Get FP parameter
            jsr         (a2)
            bne.s       snag
            
            subq.w      #1,d3                   One parameter only?
            bne.s       bad_param

            move.w      0(a1,a6.l),d0
            beq.s       fine                    Easy, SIN(0)=0
            
* Convert mantissa at 2(a1,a6.l) and exponent in D0 to IEEE format

convert     addi.w      #bias,d0                Bias exponent
            clr.l       -(a7)
            subq.l      #4,a7                   No need to clear
            clr.l       -(a7)
            move.w      d0,(a7)                 Plug in exponent
            move.l      2(a1,a6.l),d0
            bpl.s       get_mant

* Adjust sign and renormalise if necessary

            bset        #7,(a7)
            neg.l       d0
            bpl.s       get_mant

            addq.w      #1,(a7)                 Scale exponent up

* Why shift right then left? Skip the twaddle (TRAPF does it fast)
*
*           lsr.l       #1,d0                   Scale mantissa down

            SkipWord

get_mant    add.l       d0,d0                   Normalise
            move.l      d0,4(a7)

            fmove.x     (a7),fp0
            bra         call_fpsp               Compute and return

*******************************************************************
*
* Replacement SuperBASIC exponential function IEEE format FPU maths

fexp        movea.w     ($114).w,a2             Get FP parameter
            jsr         (a2)
            bne.s       fail

            subq.w      #1,d3                   One at a time
            bne.s       oops

* Check input is within a plausible range, to avert overflow.
* +/-512 is the JS ROM range yielding about 1E+/-222 respectively.
* In theory Qdos supports EXP(+/-1418) = 1E+/-615 without overflow
* but the old limit was retained for 'compatibility' until George
* observed that SMS/Q allows more. So for now it's -1024 to +1023,
* which is not quite the same as either system but allows very big
* results, albeit not quite as big as theoretically possible. If
* this 'restriction' bothers you, please suggest a fix to Simon!

            cmpi.w      #$080B,(a1,a6.l)        FIX! Was $8098000.L
            bcc.s       overflow

* Convert number on RI stack to IEEE format

            move.w      0(a1,a6.l),d0
            beq         one                     EXP(0)=1

            move.l      Exp,a4                  Find routine
            addi.w      #bias,d0                Bias exponent
            clr.l       -(a7)
            subq.l      #4,a7                   No need to clear
            clr.l       -(a7)
            move.w      d0,(a7)                 Plug in exponent
            move.l      2(a1,a6.l),d0
            bpl.s       normal

* Adjust sign and renormalise mantissa if necessary

            bset        #7,(a7)
            neg.l       d0
            bpl.s       normal

            addq.w      #1,(a7)                 Scale exponent up
            SkipWord                            Skip mantissa shift

normal      add.l       d0,d0
            move.l      d0,4(a7)

            jsr         (a4)        
            bra.s       unsigned                Return result >=0

*******************************************************************
*
* Many (but not trigonometrical parameter) errors exit here

oops        moveq       #-15,d0                 Bad parameter code
fail        rts                                 Error code in D0

overflow    moveq       #-18,d0                 Negative root?!
            rts                                 FIX! Indent RTS

*******************************************************************
*
* Replacement SuperBASIC square root function IEEE format FPU maths 
*
* This takes advantage of special properties of the SQRT function.
*
* (a) All real inputs and results can only be positive values.
* (b) SQRT(0)=0, so no need to convert or compute in this case.
* (c) SQRT of any denormalised number is zero.
*
* It converts with BFINS via 96 bit IEEE format to avoid range
* errors with single or double precision, and minimise shifting.
* This idea comes from the Thor 21 firmware by David Oliver. It
* does not round to take account of the 32nd (discarded) bit of
* the result. 30 bit accuracy is expected to be enough...

fsqrt       movea.w     ($114).w,a2             Get FP parameter
            jsr         (a2)
            bne.s       fail

            subq.w      #1,d3                   One at a time
            bne.s       oops

* Convert number on RI stack to IEEE format on A7

            move.w      0(a1,a6.l),d0
            beq         done                    SQRT(0)=0

            move.l      2(a1,a6.l),d1
            bmi         overflow                Must be positive
            
            addi.w      #bias,d0                Bias exponent
            clr.l       -(a7)

            add.l       d1,d1
            move.l      d1,-(a7)

*           subq.l      #4,a7                   No need to clear
*           bfins       d1,(a7){0:31}           Plug in mantissa

            clr.l       -(a7)
            move.w      d0,(a7)                 Plug in exponent

* Perform the operation in IEEE format 

            fmove.x     (a7),fp0
            fsqrt       fp0
unsigned    move.w      #$7FFF,d0               Mask
            fmove.x     fp0,(a7)

* Convert positive result from IEEE to Qdos format
            
            and.w       (a7),d0                 Exponent
            beq.s       zero
            move.l      4(a7),d1
            lsr.l       #1,d1                   Qdos positive value

* Perform rounding of the 32nd bit, just to be pedantic...

            bcc.s       do_exponent

            addq.l      #1,d1                   Round
            bcc.s       do_exponent

            roxr.l      #1,d1                   Divide mantissa by 2
            addq.w      #1,d0                   Double exponent

*******************************************************************
*
* Commonly-used exit points

do_exponent subi.w      #bias,d0                Adjust exponent
            bmi.s       zero                    Minimal

stack       move.w      d0,0(a1,a6.l)
            move.l      d1,2(a1,a6.l)

tidy        lea.l       12(a7),a7
done        moveq       #2,d4                   FP result
            moveq       #0,d0
            rts

zero        clr.w       0(a1,a6.l)              Return FP 0
            clr.l       2(a1,a6.l)
            bra.s       tidy

*******************************************************************
*
* Natural logarithm - input must be positive and greater than zero
*                     result is positive if input >=1 (1 returns 0)

fln         move.l      Ln,a4
flogs       movea.w     ($114).w,a2             Get FP parameter
            jsr         (a2)
            bne.s       fail

            subq.w      #1,d3                   One at a time
            bne         oops

* Convert number on RI stack to IEEE format on A7

            move.l      2(a1,a6.l),d1
            ble         overflow                Must be more than zero

            move.w      0(a1,a6.l),d0
            clr.l       -(a7)

            add.l       d1,d1
            move.l      d1,-(a7)

            addi.w      #bias,d0                Bias exponent

*           subq.l      #4,a7                   No need to clear
*           bfins       d1,(a7){0:31}           Plug in mantissa

            clr.l       -(a7)
            move.w      d0,(a7)                 Plug in exponent

* IEEE value is on A7 stack. Call appropriate FP library routine

call_fpsp   jsr         (a4)

* Return the signed IEEE value in FP0 to space on the Qdos RI stack

signed_ieee move.w      #$7FFF,d0
            fmove.x     fp0,(a7)                Re-use space
            and.w       (a7),d0                 Get exponent
            beq.s       zero

            move.l      4(a7),d1                Get mantissa
            lsr.l       #1,d1                   Normalise
            bcc.s       sign_check

            addq.l      #1,d1                   Round
            bvc.s       sign_check              **1.4** Check sign

            roxr.l      #1,d1                   Divide mantissa by 2
            addq.w      #1,d0                   Double exponent
sign_check  tst.b       (a7)                    Negative result?
            bpl.s       do_exponent

            neg.l       d1                      Reverse Qdos sign
            bra.s       do_exponent

*******************************************************************
*
* Logarithm to base 10 - input must be positive and more than zero
*                     result is positive if input >=1 (1 returns 0)

flog10      move.l      Log10,a4
            bra.s       flogs

*******************************************************************
*
* Dummy FP function for comparative timings, suggested by Thierry

            if          FNAMES=1

fnull       movea.w     ($114).w,a2             Get FP parameter
            jsr         (a2)

            subq.w      #1,d3                   One at a time
            bne         oops

            moveq       #2,d4                   FP result on RI
            rts                                 Vector result in D0

            endc        FNAMES

*******************************************************************
*
* SuperBASIC function definitions
*
* Names can either be distinct or can use SuperBASIC ones (with no
* F prefix) after successful testing, replacing "JS" & later code.

define      dc.w        0,0         No procedures

            if          FNAMES=1    True for test versions

            dc.w        13          Number of functions

            dc.w        fnull-*
            dc.b        5,'FNULL'

            dc.w        fln-*
            dc.b        3,'FLN'

            dc.w        fexp-*
            dc.b        4,'FEXP'

            dc.w        flog10-*
            dc.b        6,'FLOG10'

            dc.w        fsqrt-*
            dc.b        5,'FSQRT'

            dc.w        fsine-*
            dc.b        4,'FSIN'
            
            dc.w        fcosine-*
            dc.b        4,'FCOS'

            dc.w        farccosine-*
            dc.b        5,'FACOS'

            dc.w        farcsine-*
            dc.b        5,'FASIN'

            dc.w        ftangent-*
            dc.b        4,'FTAN'

            dc.w        farctangent-*
            dc.b        5,'FATAN'

            dc.w        fcotangent-*
            dc.b        4,'FCOT'

            dc.w        facot-*
            dc.b        5,'FACOT'

            else                    Alias Sinclair versions

            dc.w        12          Number of functions

            dc.w        facot-*
            dc.b        4,'ACOT'

            dc.w        fln-*
            dc.b        2,'LN'

            dc.w        fexp-*
            dc.b        3,'EXP'

            dc.w        flog10-*
            dc.b        5,'LOG10'

            dc.w        fsqrt-*
            dc.b        4,'SQRT'

            dc.w        fsine-*
            dc.b        3,'SIN'
            
            dc.w        fcosine-*
            dc.b        3,'COS'

            dc.w        farccosine-*
            dc.b        4,'ACOS'

            dc.w        farcsine-*
            dc.b        4,'ASIN'

            dc.w        ftangent-*
            dc.b        3,'TAN'

            dc.w        farctangent-*
            dc.b        4,'ATAN'

            dc.w        fcotangent-*
            dc.b        3,'COT'

            end_if                              NOT FNAMES

            dc.w        0
            
            end

